from PIL import Image
import requests
from io import BytesIO
import numpy as np
import pickle
import os

class FlagUtil:

    def __init__(self):
        '''
        Initializes a FlagUtil object with a color dictionary to be used for memoization.
        It stores RGB colors as keys and their respective 'web safe' colors as the values.
        This speeds up the process for converting flags to 'web safe' colors. It is also
        initialized with a list of all 216 'web safe' colors.
        '''
        self.color_dict = {}
        self.web_safe_colors = self.get_web_safe_colors()

    def pickle_numpy(self, arr, country_filename):
        '''
        Loads the given numpy array into a file with the given file name.

        Parameters
        ----------
        arr : array
            A 2D array representing an image of a country's flag

        country_filename : str
            The name of the file used to store the image of the country's flag
        '''
        with open('flags/' + country_filename + '.pkl','wb') as f:
            pickle.dump(arr, f)

    def get_web_safe_colors(self):
        '''
        Returns a list of all 216 'web safe' colors.

        Returns
        -------
        list
            A list of all 216 'web safe' colors
        '''
        lst = []
        for i in range(0, 256, 51):
            for j in range(0, 256, 51):
                for k in range(0, 256, 51):
                    tup = (i, j, k)
                    lst.append(tup)
                    
        return lst

    def find_nearest_color(self, color):
        '''
        Returns the 'web safe' color that is closest to the given
        RGB color.

        Parameters
        ----------
        color : tuple
            The color

        Returns
        -------
        tuple
            A tuple of RGB values of the 'web safe' color that is 
            most similar to the given color
        '''
        # if this color's been processed before, get its stored value
        if color in self.color_dict:
            return self.color_dict[color]
                
        min_dist = 1000
        nearest_color = (0, 0, 0)
        for cur_col in self.web_safe_colors:
            # summing up the difference of red, green, and blue values of the two colors
            dist = sum(tuple(map(lambda i, j: abs(i - j), color, cur_col)))
            if dist < min_dist:
                min_dist = dist
                nearest_color = cur_col
                
        # adding the color to the dict for future use
        self.color_dict[color] = nearest_color
        return nearest_color

    def fix_image_color(self, img):
        '''
        Standardizes the given image's colors by changing every 
        pixel's color to its closest 'web safe' color.

        Parameters
        ----------
        img : Image
            The image whose colors are being standardized
        '''
        width, height = img.size

        for x in range(width):
            for y in range(height):
                # get the current color of the pixel
                # and replace it with the 'web safe' version
                current_color = img.getpixel((x, y))
                new_color = self.find_nearest_color(current_color)
                img.putpixel((x, y), new_color)

    def process_img(self, url):
        '''
        Loads in an image from the given url and then converts it to RGB,
        standardizes its colors, and resizes it appropriately.

        Parameters
        ----------
        url : str
            A url linking to an image to be processed

        Returns
        -------
        array
            A numpy array representing the processed image from the
            given url
        '''
        # getting the image from the url
        response = requests.get(url)
        image_bytes = BytesIO(response.content)

        try:
            img = Image.open(image_bytes)
        except:
            raise IOError("Unable to open up image")
        
        # converting it to RGB
        img = img.convert("RGB")
        
        # resizing to 180 x 90
        img = img.resize((180, 90))
        
        # standardizing the colors to 'web safe' colors
        self.fix_image_color(img)
        
        # converting an array of pixels
        img = np.array(img)
        return img

    def makeArray(self, npy_file):
        '''
        Loads in a file of the given file name and returns the numpy array
        that was stored in the file.

        Parameters
        ----------
        npy_file : str 
            The name of the file being read

        Returns
        -------
        array
            A numpy array that was being stored in the given file
        '''
        with open(os.path.join(os.path.dirname(__file__), npy_file), 'rb') as npy:
            arr = pickle.load(npy)
            return arr




from .flag_identifier import FlagIdentifier

# the identifier
id = FlagIdentifier()

def identify(url, method = "mse"):
    '''
    Uses the given method (either mse, ssim, or hash) to identify the flag that is most similar
    to the image linked to the provided url.

    Parameters
    ----------
    url : str
        The url linking to an image of a flag to be identified

    method : str
        The method (one of mse, ssim, or hash) used to identify the flag that the 
        image is representing

    Returns
    -------
    str
        A string of the country name whose flag is most similar to the given flag image
    '''
    return id.identify(url, method = method)

def closest_flag(country, method = "mse"):
    '''
    Returns the name of the country whose flag is most similar to the given 
    country.

    Parameters
    ----------
    country : str
        The name of the country
    
    method : str
        The method (one of mse, ssim, or hash) used to find the flag that is
        most similar to the one of the given country

    Returns
    -------
    str
        A string of the country name whose flag is most similar to that of the given country
        name
    '''
    return id.closest_flag(country, method = method)

def farthest_flag(country, method = "mse"):
    '''
    Returns the name of the country whose flag is least similar to the given
    country.

    Parameters
    ----------
    country : str
        The name of the country

    method : str
        The method (one of mse, ssim, or hash) used to find the flag that is 
        least similar to the one of the given country

    Returns
    -------
    str
        A string of the country name whose flag is least similar to that of the given country
        name
    '''
    return id.farthest_flag(country, method = method)

def display(country):
    '''
    Displays the flag of the given country.

    Parameters
    ----------
    country : str
        The name of the country whose flag is to be displayed
    '''
    id.display(country)

def get_flag_img(country):
    '''
    Returns an Image object representing the flag of the given country name.

    Parameters
    ----------
    country : str
        The name of the country

    Returns
    -------
    Image
        an image object representing the flag of the given country name.
    '''
    return id.get_flag_img(country)

def get_flag_df():
    '''
    Returns a DataFrame of all the country names and their corresponding flags.

    Returns
    -------
    DataFrame
        A DataFrame of all the country names and their corresponding flags
    '''
    return id.get_flag_df()

def get_country_list():
    '''
    Returns a list of all 195 country names.

    Returns
    -------
    list
        A list of all the country names
    '''
    return id.get_country_list()

def flag_dist(countryA, countryB, method = "mse"):
    '''
    Gets the distance between the flags of the two given countries
    using the given method (one of mse, ssim, or hash).

    Parameters
    ----------
    countryA : str
        The name of the first country

    countryB : str
        The name of the second country

    method : str
        The method (one of mse, ssim, or hash) used to find distance between the
        flags of the two given countries. For both mse and hash, smaller values mean higher 
        similarity and a value of 0 means the two flags are identical. For ssim, higher values mean 
        higher similarity and the value must be between -1 and 1. A value of 1 for ssim
        means the two flags are identical.

    Returns
    -------
    float
        The distance between the two flags of the two given countries
    '''
    return id.flag_dist(countryA, countryB, method)


from setuptools import setup, find_packages

def readme():
    with open("README.md") as f:
        return f.read()

setup(
    name = "flagpy",
    version = "1.0.3",
    description = "Country Flag Classifier.",
    long_description = readme(),
    long_description_content_type = "text/markdown",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3"
    ],
    url = "https://github.com/saahilkumar/world-flag-identifier",
    author = "Saahil Kumar",
    author_email = "kumar.saa@northeastern.edu",
    license = "MIT",
    packages = ["flagpy"],
    package_data = {"": ["flag_df.csv", "flags/*.pkl"]},
    include_package_data = True,
    install_requires = [
        "Pillow", 
        "requests",
        "numpy",
        "pandas",
        "ImageHash",
        "scikit-image"
    ]
)


import pandas as pd
from PIL import Image
import numpy as np
from skimage import metrics
import imagehash
from .flag_util import FlagUtil
import operator
import os

class FlagIdentifier:

    def __init__(self):
        '''
        Initializes a FlagIdentifier object that has a FlagUtil object and DataFrame of 
        countries and their flags.
        '''
        self.util = FlagUtil()
        
        # reading in a df of file names which store numpy arrays
        self.flag_df = pd.read_csv(os.path.join(os.path.dirname(__file__), "flag_df.csv"), index_col = "country")

        # converting the files to numpy arrays
        self.flag_df["flag"] = self.flag_df["flag"].apply(self.util.makeArray)

    def get_flag_df(self):
        '''
        Returns a copy of this FlagIdentifier's flag DataFrame.

        Returns
        -------
        DataFrame
            A copy of this FlagIdentifier's flag DataFrame
        '''
        return self.flag_df.copy()

    def get_country_list(self):
        '''
        Returns a list of all 195 current countries stored in this FlagIdentifier's DataFrame

        Return
        ------
        list
            A list of all the countries in this FlagIdentifier's DataFrame
        '''
        return list(self.get_flag_df().index.values)

    def display(self, country):
        '''
        Displays the flag of the given country.

        Parameters
        ----------
        country : str
            The country whose flag is to be displayed
        '''
        self.get_flag_img(country).show()

    def get_flag_img(self, country):
        '''
        Returns an Image object of the flag of the given country.

        Parameters
        ----------
        country : str
            The country whose flag is returned

        Returns
        -------
        Image
            An Image object of the flag of the given country
        '''
        return Image.fromarray(self.flag_df["flag"].loc[country.title()])

    def flag_dist(self, countryA, countryB, method = "mse"):
        '''
        Uses the given method (one of mse, ssim, or hash) to find the distance
        between the two flags of the two given countries.

        Parameters
        ----------
        countryA : str
            The name of the first country

        countryB : str
            The name of the second country

        method : str
            The method (one of mse, ssim, or hash) used to find distance between the
            flags of the two given countries. For both mse and hash, smaller values mean higher 
            similarity and a value of 0 means the two flags are identical. For ssim, higher values mean 
            higher similarity and the value must be between -1 and 1. A value of 1 for ssim
            means the two flags are identical.

        Returns
        -------
        float
            The distance between the two flags of the two given countries
        '''
        flagA = self.flag_df["flag"].loc[countryA.title()]
        flagB = self.flag_df["flag"].loc[countryB.title()]
        if method == "mse":
            return self.__mse(flagA, flagB)
        elif method == "ssim":
            return self.__ssim(flagA, flagB)
        elif method == "hash":
            return self.__hash(flagA, flagB)
        else:
            raise ValueError("method must be one of: mse, ssim, hash")

    def __mse(self, imageA, imageB):
        '''
        Returns the mean-squared error between the two given images. Lower values mean
        a higher similarity between the two images (0 is perfect similarity).

        Parameters
        ----------
        imageA : array
            A numpy array representing the first image

        imageB : array
            A numpy array representing the second image

        Returns
        -------
        float
            The mean-squared error between the two given images
        '''
        # credit to: https://www.pyimagesearch.com/2014/09/15/python-compare-two-images/
        err = np.sum((imageA.astype("float") - imageB.astype("float")) ** 2)
        err /= float(imageA.shape[0] * imageA.shape[1])

        return err

    def __hash(self, imageA, imageB):
        '''
        Returns the hash distance between the two given images. Lower values mean
        a higher similarity between the two images (0 is perfect similarity).

        Parameters
        ----------
        imageA : array
            A numpy array representing the first image

        imageB : array
            A numpy array representing the second image

        Returns
        -------
        int
            The hash distance between the two given images
        '''
        firsthash = imagehash.average_hash(Image.fromarray(imageA))
        otherhash = imagehash.average_hash(Image.fromarray(imageB))
        return firsthash - otherhash

    def __ssim(self, imageA, imageB):
        '''
        Returns the structural similarity index measure (ssim) 
        between the two given images. Higher values mean
        a higher similarity between the two images (1 is perfect similarity).

        Parameters
        ----------
        imageA : array
            A numpy array representing the first image

        imageB : array
            A numpy array representing the second image

        Returns
        -------
        float
            The structural similarity index measure (ssim) between the two given images
        '''
        return metrics.structural_similarity(imageA, imageB, multichannel=True)


    def closest_flag(self, country, method = "mse"):
        '''
        Returns the name of the country whose flag is most similar to the flag
        of the given country. Uses the given method to find the closest flag.

        Parameters
        ----------
        country : str
            The name of the country

        method : str
            The method (one of mse, ssim, or hash) used to find the flag that is
            most similar to that of the given country

        Returns
        -------
        str
            The name of the country whose flag is most similar to that of the given
            country
        '''
        if method == "mse" or method == "hash":
            return self.__abstract_compare_flags(country.title(), operator.lt, method)
        elif method == "ssim":
            return self.__abstract_compare_flags(country.title(), operator.gt, method)
        else:
            raise ValueError("method must be one of: mse, ssim, or hash")


    def farthest_flag(self, country, method = "mse"):
        '''
        Returns the name of the country whose flag is least similar to the flag
        of the given country. Uses the given method to find the farthest flag.

        Parameters
        ----------
        country : str
            The name of the country

        method : str
            The method (one of mse, ssim, or hash) used to find the flag that is
            least similar to that of the given country

        Returns
        -------
        str
            The name of the country whose flag is least similar to that of the given
            country
        '''
        if method == "mse" or method == "hash":
            return self.__abstract_compare_flags(country.title(), operator.gt, method)
        elif method == "ssim":
            return self.__abstract_compare_flags(country.title(), operator.lt, method)
        else:
            raise ValueError("method must be one of: mse, ssim, or hash")

    def __abstract_compare_flags(self, country, op, method):
        '''
        Finds the country whose flag is "closest" (according to the given operation)
        to the flag of the given country. Uses the given method to calculate the distance
        between flags.

        Parameters
        ----------
        country : str
            The name of the country

        op : operator
            The operator used to find the "closest" flag

        method : str
            The method (one of mse, ssim, or hash) used to find the distance between
            flags

        Returns
        -------
        str
            The name of the country whose flag is "closest" to the flag of the given
            country
        '''
        best_dist = -1
        max_country = 0
        for c in self.flag_df.index:
            dist = self.flag_dist(country, c, method = method)
            if (op(dist, best_dist) or best_dist == -1) and c != country:
                best_dist = dist
                max_country = c
                
        return max_country

    def identify(self, url, method = "mse"):
        '''
        Returns the name of the country whose flag is most similar to the flag 
        in the image represented by the given url.

        Parameters
        ----------
        url : str
            The url that links to an image of a flag to be identified

        method : str
            The method (one of mse, ssim, or hash) used to find the flag that 
            is most similar to the one in the image of the given url

        Returns
        -------
        str
            The name of the country whose flag is most similar to the one 
            represented by the url
        '''
        if method == "mse":
            return self.__abstract_identify(url, self.__mse, operator.lt)
        elif method == "ssim":
            return self.__abstract_identify(url, self.__ssim, operator.gt)
        elif method == "hash":
            return self.__abstract_identify(url, self.__hash, operator.lt)
        else:
            raise ValueError("method must one of: mse, hash, ssim")

    def __abstract_identify(self, url, dist_func, op):
        '''
        Returns the name of the country whose flag is most similar to the
        image represented by the url. Uses the given dist_func and operator to find the 
        closest flag.

        Parameters
        ----------
        url : str
            The url that links to an image of a flag to be identified

        dist_func : function
            A distance function (either mse, ssim, or hash) to find the closest
            flag to the one in the url

        op : operator
            The operator used to find the flag closest to the one represented
            by the url
        
        Returns
        -------
        str
            The name of the country whose flag is most similar to the one in the 
            image of the url
        '''
        flag = self.util.process_img(url)
        
        best_dist = -1
        closest_index = 0
        for c in self.flag_df.index:
            cur_flag = self.flag_df["flag"].loc[c]
            dist = dist_func(flag, cur_flag)

            if op(dist, best_dist) or best_dist == -1:
                best_dist = dist
                closest_index = c
                
        return closest_index